class Mock(object):
    def __init__(self):
        self.item_history = list()
        self.attribute_history = list()
        self.call_history = list()

    def __repr__(self):
        return "Mock -- Item History: %s, Attribute History: %s, Call History: %s" % (
            str(self.item_history), str(self.attribute_history), str(self.call_history))

    def __call__(self, *args, **kwargs):
        self.call_history.append((args, kwargs))
        return self

    def __getitem__(self, name):
        self.item_history.append(name)
        return self

    def __getattr__(self, name):
        self.attribute_history.append(name)
        return self



import os
import sys

import autograder.api.courses.assignments.submissions.fetch.user.attempts
import autograder.cli.courses.assignments.submissions.common

def run(arguments):
    result = autograder.api.courses.assignments.submissions.fetch.user.attempts.send(arguments,
            exit_on_error = True)

    if (not result['found-user']):
        print("No matching user found.")
        return 1

    if (len(result['grading-results']) == 0):
        print("No attempts found.")
        return 2

    assignment = result['grading-results'][0]['info']['assignment-id']
    user = result['grading-results'][0]['info']['user']
    out_dir = os.path.join(arguments.out_dir, assignment, user)

    for grading_result in result['grading-results']:
        autograder.cli.courses.assignments.submissions.common.output_grading_result(grading_result,
                out_dir, True)

    print("Wrote %d attempts to '%s'." % (len(result['grading-results']), arguments.out_dir))

    return 0

def main():
    return run(_get_parser().parse_args())

def _get_parser():
    parser = autograder.api.courses.assignments.submissions.fetch.user.attempts._get_parser()

    parser.add_argument('-o', '--out-dir', dest = 'out_dir',
        action = 'store', type = str, default = '.',
        help = ('Where to create a new directory that will contain submission attempts.'
            + ' An existing subdirectory will be removed.'
            + ' Defaults to the current directory.'))

    return parser

if (__name__ == '__main__'):
    sys.exit(main())



API_VERSION = 'v03'

API_REQUEST_JSON_KEY = 'content'

API_RESPONSE_KEY_SUCCESS = 'success'
API_RESPONSE_KEY_MESSAGE = 'message'
API_RESPONSE_KEY_CONTENT = API_REQUEST_JSON_KEY

SERVER_ROLES = [
    'user',
    'creator',
    'admin',
    'owner',
]

COURSE_ROLES = [
    'unknown',
    'other',
    'student',
    'grader',
    'admin',
    'owner',
]



import autograder.api.constants

# Common CLI arguments.

def add_new_course_role_argument(parser, action):
    parser.add_argument('--new-course-role', dest = 'new-course-role',
        action = 'store', type = str, default = 'student',
        choices = autograder.api.constants.COURSE_ROLES,
        help = f'The course role of the user to {action} (default: %(default)s).')

def add_new_email_argument(parser, action):
    parser.add_argument('--new-email', dest = 'new-email',
        action = 'store', type = str, required = True,
        help = f'The email of the user to {action}.')

def add_new_lms_id_argument(parser, action):
    parser.add_argument('--new-lms-id', dest = 'new-lms-id',
        action = 'store', type = str, default = '',
        help = f'The lms id of the user to {action}.')

def add_new_name_argument(parser, action):
    parser.add_argument('--new-name', dest = 'new-name',
        action = 'store', type = str, default = '',
        help = f'The name of the user to {action}.')

def add_skip_emails_argument(parser):
    parser.add_argument('--skip-emails', dest = 'skip-emails',
        action = 'store_true', default = False,
        help = 'Skip sending any emails. Be aware that this may result in inaccessible'
        + ' information (default: %(default)s).')

def add_table_argument(parser):
    parser.add_argument('--table', dest = 'table',
        action = 'store_true', default = False,
        help = 'Output the results as a TSV table with a header (default: %(default)s).')



# Read a TSV file and return a list of lists of the stripped fields in the TSV file.
# Raise an error if a line has more fields than the maximum length
# or less fields than the minimum length.
def load_tsv(path, max_len, min_len = 1):
    rows = []

    with open(path, 'r') as file:
        lineno = 0
        for line in file:
            lineno += 1

            line = line.strip()

            if (line == ""):
                continue

            row = line.split("\t")
            row = [field.strip() for field in row]

            if (len(row) < min_len):
                raise ValueError(
                    "File ('%s') line (%d) has too few values. Min is %d, found %d." % (
                        path, lineno, min_len, len(row)))

            if (len(row) > max_len):
                raise ValueError(
                    "File ('%s') line (%d) has too many values. Max is %d, found %d." % (
                        path, lineno, max_len, len(row)))

            rows.append(row)

    return rows



import gzip
import base64

DEFAULT_ENCODING = 'utf-8'

def from_base64(contents, encoding = DEFAULT_ENCODING, **kwargs):
    contents = base64.b64decode(contents.encode(encoding), validate = True)
    return from_bytes(contents, encoding = encoding, **kwargs)

def from_bytes(content, string = True, encoding = DEFAULT_ENCODING):
    data = gzip.decompress(content)

    if (not string):
        return data

    return data.decode(encoding)



import sys

import autograder.api.courses.admin.update

def run(arguments):
    result = autograder.api.courses.admin.update.send(arguments, exit_on_error = True)

    if (result['course-updated']):
        print("Course updated.")
    else:
        print("Course not updated.")
        print("The request to the server was successful, no errors were encountered.")
        print("Does the course have a source?")

    return 0

def main():
    return run(_get_parser().parse_args())

def _get_parser():
    return autograder.api.courses.admin.update._get_parser()

if (__name__ == '__main__'):
    sys.exit(main())



import sys

import autograder.api.courses.assignments.submissions.submit
import autograder.assignment
import autograder.util.timestamp

def run(arguments):
    result = autograder.api.courses.assignments.submissions.submit.send(arguments, arguments.files,
            exit_on_error = True)

    message = result.get('message', '')
    if ((message is not None) and (message != '')):
        # Replace any timestamps in the message.
        message = autograder.util.timestamp.convert_message(message, pretty = True)

        print("--- Message from Autograder ---")
        print(message)
        print("-------------------------------")

    if (result['rejected']):
        print("Submission was rejected by the autograder.")
        return 1

    if (not result['grading-success']):
        print("Grading failed.")
        return 2

    submission = autograder.assignment.GradedAssignment.from_dict(result['result'])
    print(submission.report())
    return 0

def main():
    return run(_get_parser().parse_args())

def _get_parser():
    parser = autograder.api.courses.assignments.submissions.submit._get_parser()

    parser.add_argument('files', metavar = 'FILE',
        action = 'store', type = str, nargs = '+',
        help = 'The path to your submission file(s).')

    return parser

if (__name__ == '__main__'):
    sys.exit(main())



import http
import sys

import autograder.api.error
import autograder.api.users.auth

def run(arguments):
    try:
        return _auth(arguments)
    except autograder.api.error.AutograderError as ex:
        print("ERROR: " + ex.args[0], file = sys.stderr)
        return 2

def _auth(arguments):
    try:
        autograder.api.users.auth.send(arguments, exit_on_error = False)
    except autograder.api.error.APIError as ex:
        if (ex.code != http.HTTPStatus.UNAUTHORIZED):
            raise ex

        print("Authentication failed.")
        return 1

    print("Authentication successful.")
    return 0

def main():
    return run(_get_parser().parse_args())

def _get_parser():
    parser = autograder.api.users.auth._get_parser()
    return parser

if (__name__ == '__main__'):
    sys.exit(main())



import sys

import autograder.api.courses.assignments.submissions.fetch.user.attempt
import autograder.cli.courses.assignments.submissions.common

def run(arguments):
    result = autograder.api.courses.assignments.submissions.fetch.user.attempt.send(arguments,
            exit_on_error = True)

    if (not result['found-user']):
        print("No matching user found.")
        return 1

    if (not result['found-submission']):
        print("No matching submission found.")
        return 2

    autograder.cli.courses.assignments.submissions.common.output_grading_result(
        result['grading-result'], arguments.out_dir)

    print("Wrote submission to '%s'." % (arguments.out_dir))

    return 0

def main():
    return run(_get_parser().parse_args())

def _get_parser():
    parser = autograder.api.courses.assignments.submissions.fetch.user.attempt._get_parser()

    parser.add_argument('-o', '--out-dir', dest = 'out_dir',
        action = 'store', type = str, default = '.',
        help = ('Where to create a new directory that contains the submission information.'
            + ' An existing subdirectory will be removed.'
            + ' Defaults to the current directory.'))

    return parser

if (__name__ == '__main__'):
    sys.exit(main())



import autograder.api.common
import autograder.api.config

API_ENDPOINT = 'users/password/reset'
API_PARAMS = [
    autograder.api.config.PARAM_USER_EMAIL,
]

DESCRIPTION = 'Reset to a random password that will be emailed to you.'

def send(arguments, **kwargs):
    return autograder.api.common.handle_api_request(arguments, API_PARAMS, API_ENDPOINT, **kwargs)

def _get_parser():
    parser = autograder.api.config.get_argument_parser(
        description = DESCRIPTION,
        params = API_PARAMS)

    return parser



import requests

def get(url, dest):
    response = requests.get(url)

    if (response.status_code != 200):
        raise ValueError("Non-Success HTTP status (%d) on GET for: '%s'." % (
            response.status_code, url))

    with open(dest, 'wb') as file:
        file.write(response.content)



import os
import types

import autograder.code

ALL_SUBMISSION_KEY = '__all__'

def prepare(path, raise_on_collision = False):
    """
    Get a submission from a path, prepare it for grading,
    and return a submission namespace that contains all parsed entities.

    Directories will be fully recursively descended and each python or notebook will be prepared.

    Entries that begin with two underscores will not be included.

    The following things will be done to each file:
        1) The code will be parsed and sanitized.
        2) The sanitized code will be imported as a module into a namespace.
        3) The namespace will be made available in the returned namespace
            in a structure that matches its package structure.
            E.g. "./a/b/c.py" will be available from `prepare('.').a.b.c`,
            and "foo.py" will be available from `prepare('foo.py').foo`.
        4) All entries in the module will be put in the ALL_SUBMISSION_KEY attribute of the
            returned namespace.
            If raise_on_collision is True, an error will be raised if a key already exists.
    """

    submission = {}

    if (os.path.isfile(path)):
        _prepare_submission_file(submission, path, [], raise_on_collision)
    else:
        _prepare_submission_dir(submission, path, [], raise_on_collision)

    return _dict_to_namespace(submission)

def _prepare_submission_dir(submission, path, prefix, raise_on_collision):
    if (not os.path.isdir(path)):
        raise ValueError("Preparation target must be a dir: '%s'." % path)

    for dirent in os.listdir(path):
        dirent_path = os.path.join(path, dirent)

        if (os.path.isfile(dirent_path)):
            if (os.path.splitext(dirent)[1] not in autograder.code.ALLOWED_EXTENSIONS):
                continue

            _prepare_submission_file(submission, dirent_path, prefix, raise_on_collision)
        else:
            _prepare_submission_dir(submission, dirent_path, prefix + [dirent], raise_on_collision)

def _prepare_submission_file(submission, path, prefix, raise_on_collision):
    if (not os.path.isfile(path)):
        raise ValueError("Preparation target must be a file: '%s'." % path)

    basename = os.path.splitext(os.path.basename(path))[0]

    defs = autograder.code.sanitize_and_import_path(path, as_dict = True)

    for (name, value) in defs.items():
        if (name.startswith('__')):
            continue

        if ((name in submission) and (raise_on_collision)):
            raise ValueError("Name collision ('%s') when importing all keys for '%s'." % (
                name, path))

        if (ALL_SUBMISSION_KEY not in submission):
            submission[ALL_SUBMISSION_KEY] = {}

        submission[ALL_SUBMISSION_KEY][name] = defs[name]

    # Place the defs into a structure that matches the path.
    context = submission
    for part in prefix + [basename]:
        if ((part in context) and (not isinstance(context[part], dict))):
            raise ValueError("Name collision ('%s') when importing all keys for '%s'." % (
                part, path))

        if (part not in context):
            context[part] = {}

        context = context[part]

    context.update(defs)

def _dict_to_namespace(root):
    if (not isinstance(root, dict)):
        return root

    for (key, value) in root.items():
        root[key] = _dict_to_namespace(value)

    return types.SimpleNamespace(**root)



import ast
import sys

import autograder.api.lms.uploadscores

def run(arguments):
    arguments = vars(arguments)
    arguments['scores'] = _load_scores(arguments['path'])

    result = autograder.api.lms.uploadscores.send(arguments, exit_on_error = True)

    print("Upload complete.")
    print("    Grades Uploaded: %d" % (result['count']))
    print("    Errors: %d" % (result['error-count']))

    check_lists = [
        ['Unrecognized Users', 'unrecognized-users'],
        ['Users without LMS IDs', 'no-lms-id-users'],
    ]

    for (name, key) in check_lists:
        if ((key not in result) or (result[key] is None)):
            continue

        print("        %s: %d" % (name, len(result[key])))
        for value in result[key]:
            print("            Row %d -- %s" % (value['row'], value['entry']))

    return 0

def _load_scores(path):
    scores = []

    with open(path, 'r') as file:
        lineno = 0
        for line in file:
            lineno += 1

            line = line.strip()
            if (line == ""):
                continue

            parts = line.split("\t")
            if (len(parts) != 2):
                raise ValueError(
                    "File ('%s') line (%d) has the incorrect number of values." % (path, lineno)
                    + " Expecting 2, found %d." % (len(parts)))

            try:
                parts[1] = ast.literal_eval(parts[1])
            except Exception as ex:
                raise ValueError(
                    "File ('%s') line (%d) has a score that cannot be" % (path, lineno)
                    + " converted to a number: '%s'." % (line)) from ex

            scores.append({
                'email': parts[0],
                'score': parts[1],
            })

    return scores

def main():
    return run(_get_parser().parse_args())

def _get_parser():
    parser = autograder.api.lms.uploadscores._get_parser()

    parser.add_argument('path', metavar = 'PATH',
        action = 'store', type = str,
        help = 'Path to a TSV file with two columns: email and score.')

    return parser

if (__name__ == '__main__'):
    sys.exit(main())



import argparse
import json
import os
import sys

import autograder.submission

DEFAULT_ASSIGNMENT = 'assignment.json'
TEST_SUBMISSION_FILENAME = 'test-submission.json'

def run(args):
    assignment_config_path = os.path.abspath(args.assignment)
    submission_path = os.path.abspath(args.submission)

    grading_dir = autograder.submission.prep_grading_dir(assignment_config_path,
        submission_path, debug = args.debug)

    result = autograder.submission.run_submission(grading_dir,
            assignment_config_path = assignment_config_path)
    if (result is None):
        return 2

    print(result.report())

    if (args.out_path is not None):
        out_path = os.path.abspath(args.out_path)
        os.makedirs(os.path.dirname(os.path.abspath(out_path)), exist_ok = True)

        with open(out_path, 'w') as file:
            json.dump(result.to_dict(), file, indent = 4)

    if (args.test_submission_path is not None):
        test_submission_path = os.path.abspath(args.test_submission_path)
        if (os.path.isdir(test_submission_path)):
            test_submission_path = os.path.join(test_submission_path, TEST_SUBMISSION_FILENAME)

        os.makedirs(os.path.dirname(test_submission_path), exist_ok = True)

        with open(test_submission_path, 'w') as file:
            json.dump(result.to_test_submission(), file, indent = 4)

    return 0

def _create_test_submission(result):
    result = result.to_dict()

def _get_parser():
    parser = argparse.ArgumentParser(description =
        ('Grade an assignment (specified by an assignment JSON file) with the given submission.'
        + ' Non-Python assignments can be graded, but they require an "invocation" field'
        + ' in the assignment config, and the running machine must be configured to run them'
        + ' (e.g. have all the required software installed).'))

    parser.add_argument('-a', '--assignment',
        action = 'store', type = str, required = False, default = DEFAULT_ASSIGNMENT,
        help = 'The path to a JSON file describing an assignment (default: %(default)s).')

    parser.add_argument('-s', '--submission',
        action = 'store', type = str, required = True,
        help = 'The path to a submission to use for grading.')

    parser.add_argument('-o', '--out-path', dest = 'out_path',
        action = 'store', type = str, required = False, default = None,
        help = 'The path to a output the JSON result.')

    parser.add_argument('-t', '--test-submission-path', dest = 'test_submission_path',
        action = 'store', type = str, required = False, default = None,
        help = 'Create a test submission file at the specified path.'
            + ' If an existing dir is provided,'
            + ' a \'%s\' file will be created inside that dir.' % TEST_SUBMISSION_FILENAME)

    parser.add_argument('-d', '--debug', dest = 'debug',
        action = 'store_true', default = False,
        help = 'Enable additional output and leave behind artifacts (default: %(default)s).')

    return parser

def main():
    return run(_get_parser().parse_args())

if (__name__ == '__main__'):
    sys.exit(main())



import atexit
import os
import shutil
import tempfile
import uuid

def get_temp_path(prefix = '', suffix = '', rm = True):
    """
    Get a path to a valid (but not currently existing) temp dirent.
    If rm is True, then the dirent will be attempted to be deleted on exit
    (no error will occur if the path is not there).
    """

    path = None
    while ((path is None) or os.path.exists(path)):
        path = os.path.join(tempfile.gettempdir(), prefix + str(uuid.uuid4()) + suffix)

    if (rm):
        atexit.register(remove, path)

    return path

def remove(path):
    if (not os.path.exists(path)):
        return

    if (os.path.isfile(path) or os.path.islink(path)):
        os.remove(path)
    elif (os.path.isdir(path)):
        shutil.rmtree(path)
    else:
        raise ValueError("Unknown type of dirent: '%s'." % (path))

def copy(source, dest):
    """
    Copy a file or directory into dest.
    If source is a file, then dest can be a file or dir.
    If source is a dir, then dest must be a non-existent dir.
    """

    if (os.path.isfile(source)):
        shutil.copy2(source, dest)
    else:
        shutil.copytree(source, dest)

def copy_contents(source, dest):
    """
    Copy a file or the contents of a directory (excluding the top-level directory) into dest.
    For a file: `cp source dest/`
    For a dir: `cp -r source/* dest/`
    """

    source = os.path.abspath(source)

    if (os.path.isfile(source)):
        shutil.copy2(source, dest)
        return

    for dirent in os.listdir(source):
        source_path = os.path.join(source, dirent)
        dest_path = os.path.join(dest, dirent)

        if (os.path.isfile(source_path)):
            shutil.copy2(source_path, dest_path)
        else:
            shutil.copytree(source_path, dest_path)



import secrets

import argon2.low_level

DEFAULT_PASSWORD_LEN = 32
SALT_LENGTH_BYTES = 16

ARGON2_KEY_LEN_BYTES = 32
ARGON2_MEM_KB = 64 * 1024
ARGON2_THREADS = 4
ARGON2_TIME = 1

ENCODING = 'utf-8'

def rand_pass():
    return rand_bytes(DEFAULT_PASSWORD_LEN // 2).hex()

def rand_bytes(length):
    return secrets.token_bytes(length)

def hash(password):
    salt = rand_bytes(SALT_LENGTH_BYTES)

    hashpass = argon2.low_level.hash_secret_raw(
        password.encode(ENCODING), salt,
        time_cost = ARGON2_TIME, memory_cost = ARGON2_MEM_KB,
        parallelism = ARGON2_THREADS, hash_len = ARGON2_KEY_LEN_BYTES,
        type = argon2.low_level.Type.ID)

    return hashpass.hex(), salt.hex()



import sys

import autograder.api.users.tokens.create

def run(arguments):
    result = autograder.api.users.tokens.create.send(arguments, exit_on_error = True)

    print("Token ID: " + result['token-id'])
    print("Token Text: " + result['token-cleartext'])
    print("")
    print("Copy down the token text and keep it safe, this will be the only time it is ever shown.")

    return 0

def main():
    return run(_get_parser().parse_args())

def _get_parser():
    parser = autograder.api.users.tokens.create._get_parser()

    return parser

if (__name__ == '__main__'):
    sys.exit(main())



import sys

import autograder.cli.courses.assignments.submissions.fetch.user.history as alias

def main():
    return alias.run(_get_parser().parse_args())

def _get_parser():
    parser = alias._get_parser()
    parser.epilog = "This is an alias for `%s`." % (alias.__name__)
    return parser

if (__name__ == '__main__'):
    sys.exit(main())



import sys

import autograder.api.constants
import autograder.api.courses.users.enroll
import autograder.cli.common
import autograder.cli.config
import autograder.util.hash
import autograder.util.load

def run(arguments):
    arguments = vars(arguments)

    arguments['raw-course-users'] = _load_users(arguments['path'])
    arguments['send-emails'] = not arguments['skip-emails']

    result = autograder.api.courses.users.enroll.send(arguments, exit_on_error = True)

    autograder.cli.common.list_user_op_responses(result['results'], table = arguments['table'])
    return 0

def _load_users(path):
    users = []

    rows = autograder.util.load.load_tsv(path, 4)
    for lineno in range(len(rows)):
        row = rows[lineno]

        email = row.pop(0)

        name = ''
        if (len(row) > 0):
            name = row.pop(0)

        course_role = ''
        if (len(row) > 0):
            course_role = row.pop(0)
            course_role = course_role.lower()

        if (course_role == ''):
            course_role = 'unknown'

        if (course_role not in autograder.api.constants.COURSE_ROLES):
            raise ValueError(
                "File ('%s') line (%d) has an invalid course role '%s'." % (
                    path, lineno, course_role))

        course_lms_id = ''
        if (len(row) > 0):
            course_lms_id = row.pop(0)

        users.append({
            'email': email,
            'name': name,
            'course-role': course_role,
            'course-lms-id': course_lms_id,
        })

    return users

def main():
    return run(_get_parser().parse_args())

def _get_parser():
    parser = autograder.api.courses.users.enroll._get_parser()

    parser.description = ('Enroll users to the course from a TSV file.')

    autograder.cli.config.add_table_argument(parser)
    autograder.cli.config.add_skip_emails_argument(parser)

    parser.add_argument('path', metavar = 'PATH',
        action = 'store', type = str,
        help = 'Path to a TSV file where each line contains up to four columns:'
                + ' [email, name, course-role, lms-id].'
                + ' Only the email is required. Leading and trailing whitespace is stripped'
                + ' from all fields.')

    return parser

if (__name__ == '__main__'):
    sys.exit(main())



import autograder.api.common
import autograder.api.config

API_ENDPOINT = 'courses/assignments/submissions/submit'
API_PARAMS = [
    autograder.api.config.PARAM_COURSE_ID,
    autograder.api.config.PARAM_USER_EMAIL,
    autograder.api.config.PARAM_USER_PASS,
    autograder.api.config.PARAM_ASSIGNMENT_ID,

    autograder.api.config.APIParam('message',
        'An optional message to attatch to the submission.',
        required = False)
]

DESCRIPTION = 'Submit an assignment submission to the autograder.'

def send(arguments, files, **kwargs):
    return autograder.api.common.handle_api_request(arguments, API_PARAMS, API_ENDPOINT,
        files = files, **kwargs)

def _get_parser():
    parser = autograder.api.config.get_argument_parser(
        description = DESCRIPTION,
        params = API_PARAMS)

    return parser



"""
Handle timestamp operations.
All datetimes/timestamps passed by the autograder are ints (milliseconds since UNIX epoch (UTC)).
We will handle timestamps the same way (with the exception of pretty strings).
"""

import datetime
import re
import time

PRETTY_TIMESTAMP_FORMAT = '%Y-%m-%d %H:%M'
UNKNOWN_TIMESTAMP = "<Unknown Time (%s)>"
MISSING_TIMESTAMP = "<Missing Time>"

MESSAGE_REGEX = r'<timestamp:(-?\d+|nil)>'

def get(source = None, pretty = False, adjust_tz = True):
    if (source == MISSING_TIMESTAMP):
        return source

    if (isinstance(source, str) and ('Unknown Time' in source)):
        return source

    timestamp, clean_source = _parse_timestamp(source)
    if (timestamp is None):
        return UNKNOWN_TIMESTAMP % (clean_source)

    if (pretty):
        return _to_string(timestamp, adjust_tz = adjust_tz)

    return timestamp

def convert_message(text, pretty = False, adjust_tz = True):
    """
    Look for any timestamps embedded in the text and replace them.
    """

    while True:
        match = re.search(MESSAGE_REGEX, text)
        if (match is None):
            break

        initial_text = match.group(0)
        timestamp = match.group(1)

        if (timestamp == 'nil'):
            replacement = MISSING_TIMESTAMP
        else:
            replacement = str(get(timestamp, pretty = pretty, adjust_tz = adjust_tz))

        text = text.replace(initial_text, replacement)

    return text

def _parse_timestamp(source = None):
    if (source is None):
        return _now(), None

    if (isinstance(source, datetime.datetime)):
        return _timestamp_from_pytime(source)

    if (source == ''):
        return None, ''

    if (isinstance(source, (int, float))):
        # Unix timestamp.
        return int(source), None

    if (not isinstance(source, str)):
        raise ValueError("Unknown type ('%s') for timestamp source." % (type(source)))

    source = source.strip()

    # Try once more for a unix timestamp.
    if (re.match(r'^-?\d+(\.\d+)?$', source)):
        return int(source), None

    # Parse out some cases that Python <= 3.10 cannot deal with.
    # This will remove fractional seconds.
    source = re.sub(r'Z$', '+00:00', source)
    source = re.sub(r'(\d\d:\d\d)(\.\d+)', r'\1', source)

    try:
        value = datetime.datetime.fromisoformat(source)
        return _timestamp_from_pytime(value), None
    except Exception:
        pass

    return None, source

def _timestamp_from_pytime(pytime):
    return int(pytime.timestamp() * 1000)

def _pytime_from_timestamp(timestamp):
    return datetime.datetime.fromtimestamp(timestamp / 1000, datetime.timezone.utc)

def _now():
    return int(time.time() * 1000)

def _to_string(timestamp, adjust_tz = True):
    pytime = _pytime_from_timestamp(timestamp)
    if (adjust_tz):
        pytime = pytime.astimezone()

    return pytime.strftime(PRETTY_TIMESTAMP_FORMAT)



import sys

import autograder.api.config
import autograder.api.courses.users.enroll
import autograder.cli.common
import autograder.cli.config

def run(arguments):
    arguments = vars(arguments)

    arguments['raw-course-users'] = [{
        'email': arguments['new-email'],
        'name': arguments['new-name'],
        'course-role': arguments['new-course-role'],
        'course-lms-id': arguments['new-lms-id'],
    }]

    arguments['send-emails'] = not arguments['skip-emails']

    result = autograder.api.courses.users.enroll.send(arguments, exit_on_error = True)

    autograder.cli.common.list_user_op_responses(result['results'], table = arguments['table'])
    return 0

def main():
    return run(_get_parser().parse_args())

def _get_parser():
    parser = autograder.api.courses.users.enroll._get_parser()

    autograder.cli.config.add_table_argument(parser)
    autograder.cli.config.add_skip_emails_argument(parser)

    autograder.cli.config.add_new_email_argument(parser, "enroll")
    autograder.cli.config.add_new_name_argument(parser, "enroll")
    autograder.cli.config.add_new_course_role_argument(parser, "enroll")
    autograder.cli.config.add_new_lms_id_argument(parser, "enroll")

    return parser

if (__name__ == '__main__'):
    sys.exit(main())



import sys

def main():
    print(("You have invoked the `autograder` package directly."
        + " You probably want the `autograder.cli` package instead."))
    return 0

if (__name__ == '__main__'):
    sys.exit(main())



"""
The `autograder.cli.logs` package contains tools for
managing log records from the autograder server.
"""

import sys

import autograder.util.cli

def main():
    return autograder.util.cli.main()

if (__name__ == '__main__'):
    sys.exit(main())



import os
import setuptools

THIS_DIR = os.path.abspath(os.path.dirname(os.path.realpath(__file__)))

LICENSE_PATH = os.path.join(THIS_DIR, 'LICENSE.txt')
LONG_DESCRIPTION_PATH = os.path.join(THIS_DIR, 'README.md')
REQUIREMENTS_PATH = os.path.join(THIS_DIR, 'requirements.txt')

def get_description():
    with open(LONG_DESCRIPTION_PATH, 'r') as file:
        return file.read()

setuptools.setup(
    name = 'autograder-py',
    url = 'https://github.com/eriq-augustine/autograder-py',

    version = '0.5.3',
    keywords = 'grading',

    description = "The Python interface for the autograding server.",
    long_description = get_description(),
    long_description_content_type = 'text/markdown',


    maintainer = 'Eriq Augustine',
    maintainer_email = 'eaugusti@ucsc.edu',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers = [
        'Intended Audience :: Education',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.8',
    ],

    packages = setuptools.find_packages(
        include = ["autograder*"],
        exclude = ["tests"],
    ),

    install_requires = [
        'argon2-cffi>=23.1.0',
        'flake8>=6.0.0',
        'GitPython>=3.1.31',
        'platformdirs>=3.10.0',
        'requests>=2.31.0',
    ],

    license_files = (LICENSE_PATH, ),

    python_requires = '>=3.8'
)



import sys

import autograder.api.users.get
import autograder.cli.common
import autograder.cli.config

def run(arguments):
    result = autograder.api.users.get.send(arguments, exit_on_error = True)

    if (not result['found']):
        print("User not found.")
        return 1

    autograder.cli.common.list_users([result['user']], False, table = arguments.table,
        normalize = arguments.normalize)
    return 0

def main():
    return run(_get_parser().parse_args())

def _get_parser():
    parser = autograder.api.users.get._get_parser()

    autograder.cli.config.add_table_argument(parser)

    parser.add_argument('--normalize', dest = 'normalize',
        action = 'store_true', default = False,
        help = 'Normalize the TSV table to include at most one course enrollment per line. If the'
            + ' user is enrolled in multiple courses, they will appear multiple times. Each line'
            + ' contains the following columns: [email, name, role, course-id, course-name,'
            + ' course-role]. Only applies if --table is set (default: %(default)s).')

    return parser

if (__name__ == '__main__'):
    sys.exit(main())



import sys

import autograder.api.users.tokens.delete

def run(arguments):
    result = autograder.api.users.tokens.delete.send(arguments, exit_on_error = True)

    if (not result['found']):
        print("Token not found.")
        return 1

    print("Token deleted.")
    return 0

def main():
    return run(_get_parser().parse_args())

def _get_parser():
    parser = autograder.api.users.tokens.delete._get_parser()

    return parser

if (__name__ == '__main__'):
    sys.exit(main())



import unittest

import autograder.util.timestamp

class TestTimestamp(unittest.TestCase):
    def test_timestamp_conversions(self):
        # [(input, normal, pretty), ...]
        test_cases = [
            # Full information.
            ('2023-09-28T04:00:20Z', 1695873620000, '2023-09-28 04:00'),
            ('2023-09-28T04:00:20+00:00', 1695873620000, '2023-09-28 04:00'),
            ('2023-09-28T13:10:44+00:00', 1695906644000, '2023-09-28 13:10'),

            # Fractional seconds must be removed because of older python versions.
            ('2023-09-28T04:00:20.683684Z', 1695873620000, '2023-09-28 04:00'),
            ('2023-09-28T04:00:20.683684+00:00', 1695873620000, '2023-09-28 04:00'),
            ('2023-09-28T13:10:44.432050+00:00', 1695906644000, '2023-09-28 13:10'),
            ('2023-09-28T13:10:44.43205+00:00', 1695906644000, '2023-09-28 13:10'),

            (1695873620000, 1695873620000, '2023-09-28 04:00'),
            ('1695873620000', 1695873620000, '2023-09-28 04:00'),

            # Unknown format.
            ('abc', '<Unknown Time (abc)>', '<Unknown Time (abc)>'),
        ]

        for i in range(len(test_cases)):
            (source, expected_normal, expected_pretty) = test_cases[i]

            first_normal_timestamp = autograder.util.timestamp.get(source,
                    pretty = False, adjust_tz = False)
            second_normal_timestamp = autograder.util.timestamp.get(first_normal_timestamp,
                    pretty = False, adjust_tz = False)

            self.assertEqual(expected_normal, first_normal_timestamp,
                    "Case %d: [%s] First Normal Timestamp" % (i, source))

            self.assertEqual(expected_normal, second_normal_timestamp,
                    "Case %d: [%s] Second Normal Timestamp" % (i, source))

            first_pretty_timestamp = autograder.util.timestamp.get(source,
                    pretty = True, adjust_tz = False)

            # Once created, pretty timestamps cannot be parsed (as they lack timezone information).

            self.assertEqual(expected_pretty, first_pretty_timestamp,
                    "Case %d: [%s] First Pretty Timestamp" % (i, source))

            # Converting from normal to pretty is possible, since no information is list.
            # In the other direction, timezone information is lost.
            normal_pretty_timestamp = autograder.util.timestamp.get(first_normal_timestamp,
                    pretty = True, adjust_tz = False)
            self.assertEqual(expected_pretty, normal_pretty_timestamp,
                    "Case %d: [%s] Normal Then Pretty Timestamp" % (i, source))

    def test_timestamp_message(self):
        """
        Test pulling timestamps out of messages.
        """

        # [(input, expected), ...]
        test_cases = [
            ("Some <timestamp:0> timestamp.", "Some 1970-01-01 00:00 timestamp."),
            ("Some <timestamp:1695873620000> timestamp.", "Some 2023-09-28 04:00 timestamp."),
            ("No timestamp here.", "No timestamp here."),
            ("Some <timestamp:nil> timestamp.", "Some <Missing Time> timestamp."),
            ("Some <timestamp:-60000> timestamp.", "Some 1969-12-31 23:59 timestamp."),
            ("Some <timestamp:-60001> timestamp.", "Some 1969-12-31 23:58 timestamp."),
            ("Timestamp one <timestamp:0> and two <timestamp:-60000>.",
                "Timestamp one 1970-01-01 00:00 and two 1969-12-31 23:59."),
            ("Timestamp one <timestamp:0> and two <timestamp:0>.",
                "Timestamp one 1970-01-01 00:00 and two 1970-01-01 00:00."),
        ]

        for i in range(len(test_cases)):
            (text, expected) = test_cases[i]

            actual = autograder.util.timestamp.convert_message(text,
                    pretty = True, adjust_tz = False)
            self.assertEqual(expected, actual, "Case %d" % (i))
